# Processor.py
'''To be used for handling any operational component of wheat rust early warning
system. 

Command-line options specify which component and which json configuration 
files to run, and complete any remaining elements of the configuration files.

Component varies by choices of process_pre_job(), process_in_job() and 
process_EWS_plotting().

This runs all dates of a single forecast.

Example usage::
    $ conda activate py3EWSepi
    $ python Processor.py -p Deposition --islive --config config_Nepal_template.json --noupload -s 20200715
or::
    $ ./run_Processor.sh -p Deposition --islive --config config_Nepal_template.json --noupload -s 20200715
'''
from abc import abstractmethod, ABCMeta
from argparse import ArgumentParser
from typing import List

from ews.coordinator.utils import processor_utils
from ews.coordinator.utils.jobstatus import Jobstatus
from ews.coordinator.utils.processor_utils import end_script, end_job, append_item_to_list, \
    clear_up, parse_json_file_with_tokens


import argparse
import datetime
import json
import logging
import os
from pathlib import Path
import sys


# submodules of this project

logger = logging.getLogger(__name__)


class ProcessorBase:

    __metaclass__ = ABCMeta

    def __init__(self) -> None:
        super().__init__()
        time_now = datetime.datetime.today()
        self.today_string = time_now.strftime('%Y%m%d')
        self.now_string = time_now.strftime('%Y%m%d-%H%M-%S')

    def parse_and_check_args(self) -> dict:

        # define the command line arguments
        my_parser: ArgumentParser = argparse.ArgumentParser(
            description = 'Command-line arguments for coordinator script of env suitability model')

        # Add the arguments
        # positional arguments do not start with - or -- and are always required
        # optional arguments start with - or -- and default is required = False

        my_parser.add_argument(
                '-sc', '--sys_config',
                metavar = 'path',
                type = str,
                dest = 'sys_config_path',
                required = True,
                help = '''path to the system config file. This file contains values which are common to all Processor 
                components''')

        my_parser.add_argument(
                '-c', '--config',
                metavar = 'path',
                type = str,
                dest = 'config_path',
                required = True,
                help = '''path to a config file.''')

        my_parser.add_argument(
                '-l', '--loglevel',
                action = 'store',
                choices = list(["debug", "info", "warning", "error", "critical"]),
                default = 'info',
                help = 'verbosity of log messaging (debug, info, warning, error, critical)\n default is debug',
                dest = 'log_level',  # this names the attribute that will be parsed
        )

        my_parser.add_argument(
                '--islive',
                action = 'store_true',
                help = 'If live, email messages are sent to maintainers for warning and errors',
                dest = 'live',
        )

        my_parser.add_argument(
                '-d', '--run-date',
                metavar = 'YYYYMMDD',
                action = 'store',
                default = self.today_string,
                help = 'Initial day of calculation, starting at 00 UTC (Default is today)',
                dest = 'start_date',
        )

        my_parser.add_argument(
                '--noupload',
                action = 'store_true',
                help = 'whether results of script should be saved to willow public directory'
        )

        my_parser.add_argument(
                '--clearup',
                action = 'store_true',
                help = 'whether to delete mid-process files at the end of a successful job',
                dest = 'clearup',
        )

        self.add_additional_command_line_args(my_parser)

        # get an object holding all of the args
        args = my_parser.parse_args()

        # Check the args

        print(f"Command-line options are:\n{args}")

        # check the startstring
        if args.start_date is not self.today_string:
            try:
                # check date string is formatted correctly
                provided_start_date = datetime.datetime.strptime(args.start_date, '%Y%m%d')
                today_date = datetime.datetime.strptime(self.today_string, '%Y%m%d')

                # early limit is quite arbitrary, but this is earliest year of available survey data for Ethiopia
                date_limit_early = datetime.datetime.strptime('20070101', '%Y%m%d')
                assert date_limit_early < provided_start_date
                assert provided_start_date <= today_date

            except (ValueError, AssertionError) as e:
                raise Exception(
                    "Provided start date string is formatted incorrectly or out of range, or end date not also defined")

        dictionary: dict = vars(args)
        return dictionary

    def add_additional_command_line_args(self, arg_parser: ArgumentParser) -> None:
        """
        override this method to add additional command line arguments

        :param arg_parser:
        :return:
        """
        pass

    @staticmethod
    def generate_job_directory_path(config: dict) -> str:
        """
        Generate the job directory path

        :param config:
        :return:`
        """
        short_name: str = config['short_name']
        start_date: str = config['StartString']
        workspace_path = config['WorkspacePathout']

        job_path: str = f'{workspace_path}{short_name}_{start_date}'

        return job_path

    def prepare_job_directory(self, job_path: str):
        """
        create job directory or archive if already exists (due to a rerun)

        :param job_path:
        :return:
        """
        if os.path.exists(job_path):
            archived_dir_path = f"{job_path}-ARCHIVE_{self.now_string}"
            message = f"Job path {job_path} already exists, so moving it to {archived_dir_path}"
            print(message)
            os.rename(job_path, archived_dir_path)

        Path(job_path).mkdir(parents = True, exist_ok = False)


    def run_process(self,
                    config: dict) -> None:

        job_path: str = self.generate_job_directory_path(config)
        self.prepare_job_directory(job_path)

        log_file_path = f"{job_path}/log.txt"

        """
        Now we have the job_dir defined, we can set up the logging (to log to the dir)
        """
        if 'LoggingConfigPath' not in config.keys():
            raise Exception(f"'LoggingConfigPath' not found in config (is set in the sys_config). Either use the "
                            f"default template file in configs/logger/template_log_config.json, or copy the template "
                            f"away and edit")

        log_config_path = config['LoggingConfigPath']
        is_live: bool = config["live"]
        processor_utils.setup_logging(log_file_path,
                                      log_config_path,
                                      is_live,
                                      config['log_level'])

        logger.info("==========")
        logger.info(f"Logging started at {datetime.datetime.now().strftime('%Y %b %d %H:%M:%S')}")
        logger.info(f"Universal config is\n{json.dumps(config, indent = 2)}")
        logger.info(f"Job path will be {job_path}")

        # run any checks before creating a job directory
        # if this fails, then make a note once there is a job directory
        ready = self.process_pre_job(config)

        # lock job directory
        status: Jobstatus = Jobstatus(job_path)
        with status:

            # check for a status file in job directory
            if status.had_initial_status:
                logger.info(f"Job path already exists and has status {status.status}")

                end_script(premature = status.status not in ['SUCCESS', 'INPROGRESS'])

            logger.info(f"Current status of job directory is {status.status}")

            # make a record if process_pre_job failed
            if not ready:
                logger.error(
                    f"Process_pre_job raised an error so making a record in the job file. For details, see earlier warnings log")
                status.reset('ERROR')

                end_job(status, ignore_inprogress = True, premature = False)

            logger.info('Starting to work on each configuration')

            config_path: str = config['config_path']
            logger.info(f'Working on config {config_path}')

            # provide specific case details to template config

            # from configtemplate store the base filename
            config_file_name = os.path.basename(config_path)
            config_file_name = os.path.splitext(config_file_name)[0]
            # config['ConfigFilePath'] = config_file_name

            # write the complete configuration file to job directory
            with open(f"{job_path}/{config_file_name}.json", 'w') as write_file:
                json.dump(config, write_file, indent = 4)

            # proc_description = universal_config['ProcessInJob']
            proc_description = 'ProcessInJob'
            try:
                component: str = config['component']
                proc_out: dict = self.process_in_job(job_path, status, config, component)
            except:
                logger.exception(f"Error in process_in_job")
                status.reset('ERROR')
                end_job(status, premature = True)

            # Set default case
            # This would be improved by implementing a class structure - note that the proc_out is not curently used
            # as we now deal with copying outputs elsewhere. Keeping it here as a placeholder.
            if proc_out is None:
                proc_out = {
                    'output':  None,
                    'clearup': None}

            paths_to_clear = []
            if 'clearup' in proc_out.keys():
                append_item_to_list(
                        proc_out['clearup'],
                        paths_to_clear,
                        proc_description,
                        status)

            # Run EWS-plotting command

            proc_description = 'ProcessEWSPlotting'
            try:
                self.process_post_job(job_path, config)
            except:
                logger.exception(f"Error in {proc_description}()")
                status.reset('ERROR')
                end_job(status, premature = True)

            logger.info('Finished with EWS-Plotting, appending images to list for transfer')

            logger.info(f'Finished with config {config_path}')

            status.reset('SUCCESS')

        short_name: str = config['short_name']
        clearup: bool = config['clearup']

        if status.is_success() & (clearup is True):
            logger.info('Clearing up')

            start_date: str = config['StartString']
            workspace_path = config['WorkspacePathout']
            clearup_dest_dir = f"{workspace_path}/clearup/{short_name}_{start_date}/"
            Path(clearup_dest_dir).mkdir(parents = True, exist_ok = True)

            logger.info(f"While developing, moving directories to this directory : {clearup_dest_dir}")

            clear_up(paths_to_clear, clearup_dest = clearup_dest_dir)

        end_script(premature = False)


    @staticmethod
    def build_config(args: dict,
                     config_path: str,
                     sys_config_path: str):

        """
        Builds a 'unified' dict by reading the sys and run config json files and replacing any tokens with values from
        the sys_config in the run_config. Additionally, values set in the args are used to supplement the config dict
        generated form the json.

        Note we dont use the logger in this method as it is called before the logging is set up.

        :param args:
        :param config_path:
        :param sys_config_path:
        :return:
        """
        try:
            sys_config = parse_json_file_with_tokens(sys_config_path)
            # read the run config json, and replace any tokens with values from the sys_config (such as the
            # directory root for the installation)
            config: dict = parse_json_file_with_tokens(config_path, sys_config)
            config['config_path'] = config_path

            """
            add the sys_config keys and values to the config dict (unless it is already declared in the config) 
            """
            for k, v in sys_config.items():
                if k not in config.keys():
                    config[k] = v
                else:
                    print(f"Key '{k}' already present in run config - not adding key with same name from "
                          f"the sys_config")

            """
            then add the args keys and values to the config, we will override any keys that are already 
            in the config
            """
            for k, v in args.items():
                if k in config.keys():
                    print(f"Key '{k}' already present in run config - overriding key with same name from "
                          f"the sys args")
                config[k] = v

            #  set various time values based on the start_date that is passed in on the args
            config['StartString'] = config.pop('start_date')  # we are renaming the 'start_date' args to 'StartString'
            # note that file date represents preceding 3 hours, so day's data starts at file timestamp 0300 UTC
            start_time: datetime = datetime.datetime.strptime(config['StartString'] + '03', '%Y%m%d%H')
            start_time_string: str = start_time.strftime('%Y-%m-%d-%H%M')
            config['StartTime'] = start_time_string

        except:
            print(f"Failure in opening or checking config {config_path}")
            # TODO: This case should test flagdir.jobStatus.__exit__()
            raise  # endJob('ERROR',premature=True)

        return config


    @abstractmethod
    def process_pre_job(self, args) -> bool:
        raise NotImplementedError

    @abstractmethod
    def process_in_job(self, job_path: str, status: Jobstatus, config: dict, component: str) -> dict:
        raise NotImplementedError

    @abstractmethod
    def process_post_job(self, jobPath, configjson):
        raise NotImplementedError

    def run_processor(self, component: str, shortname: str):
        try:
            args: dict = self.parse_and_check_args()
            args["component"] = component
            args["short_name"] = shortname

            sys_config_path: str = args['sys_config_path']
            config_path: str = args['config_path']
            config: dict = self.build_config(args, config_path, sys_config_path)

            self.run_process(config)

        except SystemExit as e:
            print("caught with code " + str(e.code))
            logger.info('run_process() exited')
            sys.exit(e.code)
        except:
            raise Exception('Uncaught exception in run_Process:')
