#jobstatus.py
'''
Basic flag-file handling within a provided directory.

Running with `conda activate py3EWSepi`.
'''
import os
from glob import glob

class Jobstatus:
    """
    Context manager to determine and change status of a job based on status file
    in a given directory.

    Only possible status values are 'SUCCESS', 'INPROGRESS', 'HELD', 'WARNING',
    'SUCCESS_WITH_WARNING', 'SUCCESS_WITH_ERROR' and 'ERROR'.

    Use jobStatus.reset(status) to change the current status file.

    Use jobStatus.get() to perform a fresh search for status file.

    If you don't want any status files for this directory anymore, use
    jobStatus.cleanup()

    Suggested usage (in between task commands):
    jobPath = 'testjob'
    js = jobStatus(jobPath)
    print(js.status) # None
    js.set('INPROGRESS')
    print(js.status) # 'INPROGRESS'
    print(js.statusPath) # 'testjob/STATUS_INPROGRESS'
    js.reset('WARNING')
    print(js.status) # 'WARNING'
    print(js.statusPath) # 'testjob/STATUS_WARNING'
    js.is_warning()
    status = js.get()
    js.reset('SUCCESS')
    #js.cleanup() # if success file not needed
    """

    _status_dict = {
        #-1: None,
        0: 'SUCCESS',
        1: 'INPROGRESS',
        2: 'WARNING',
        3: 'ERROR',
        4: 'HELD',
        5: 'SUCCESS_WITH_WARNING',
        6: 'SUCCESS_WITH_ERROR'
    }

    _acceptable_end_status = ['SUCCESS','ERROR','HELD','SUCCESS_WITH_WARNING','SUCCESS_WITH_ERROR']

    _default_end_status = 'ERROR'

    def __init__(self, job_path: str):

        # determine job directory

        self.job_path = job_path

        print(f"job path is {self.job_path}")
    
        if not os.path.isdir(self.job_path):
            raise FileNotFoundError

    def __enter__(self):
        
        # determine status

        # check if it already exists
        self.status = self.get()

        self.had_initial_status = self.has_status()

        # otherwise, initialise
        if not self.had_initial_status:
            print('No initial status, so setting to INPROGRESS')
            self.set('INPROGRESS')
        
        return self
    
    def __exit__(self, type, value, traceback):
    
        status = self.get()
    
        if self.had_initial_status:

            print('exiting without changing status')

            return

        if status not in self._acceptable_end_status:

            print(f"Changing status from {status} to {self._default_end_status}")

            self.reset(self._default_end_status)

    def cleanup(self):
        # remove any current status file
        status_paths = glob(f"{self.job_path}/STATUS_*")
        
        #print(f"statusPaths are\n:{statusPaths}")

        for f in status_paths:
            
            try:
                print(f' removing {f}')
                os.remove(f)
                self.status = None

            except OSError:
                pass # there is no file to remove
        
        return


    def reset(self, new_status_in):

        # ensure status is acceptable
        # for now, handle string only
        assert isinstance(new_status_in,str)
        assert new_status_in in self._status_dict.values()

        # check statusPath has already been defined
        assert hasattr(self,'status_path')

        # modify new status if a warning or error once occurred
        new_status = new_status_in
        top_statuses = ['WARNING','ERROR']
        if (self.status in top_statuses) & (new_status not in top_statuses):
            new_status = f"{new_status_in}_WITH_{self.status}"

        new_status_path = f"{self.job_path}/STATUS_{new_status}"

        if self.status_path is not None:

            print(f'Renaming {self.status_path} to {new_status_path}')

            os.rename(self.status_path, new_status_path)

        else:

            print(f'Creating {new_status_path}')

            # by running set below

        self.set(new_status)
        
        return


    def set(self, status):

        # ensure it is acceptable status
        # for now, handle string only
        assert isinstance(status,str)
        print(status)
        assert status in self._status_dict.values()

        # set file status
        self.status_path = f"{self.job_path}/STATUS_{status}"

        open(self.status_path, 'a').close()

        # set object status
        self.status = status
        return


    def get_from_file_path(self, statusPath):
        """Convert path to status string."""
        
        filename = statusPath.split('/')[-1]

        statusString = filename.split("STATUS_")[-1]

        print(statusString)
        if statusString not in self._status_dict.values():
            # Status file descriptor not recognised
            raise KeyError

        return statusString

    def get_status_path(self):
        """This might not be needed all the time."""
        
        status_files = glob(f"{self.job_path}/STATUS_*")
        
        if len(status_files) > 1:
            print('ERROR: require only one status file in job directory')
            raise AssertionError

        elif len(status_files) == 0:
            return None
        
        else:
            return status_files[0]


    def get(self, statuspath = None):

        """Fresh look for status file and set object status based on that.
        Does not modify self.status directly"""
        
        if statuspath is None:
            self.status_path = self.get_status_path()
        
        if self.status_path is None:
            status =  None
        else:
            status = self.get_from_file_path(self.status_path)

        return status

    # logical form of status
    def has_status(self):
        return (hasattr(self, 'status')) & (self.status in self._status_dict.values())

    def is_success(self):
        self.get()
        return self.status == 'SUCCESS'

    def is_inprogress(self):
        self.get()
        return self.status == 'INPROGRESS'

    def is_warning(self):
        self.get()
        return self.status == 'WARNING'

    def is_error(self):
        self.get()
        return self.status == 'ERROR'

    def is_held(self):
        self.get()
        return self.status == 'HELD'
